#include <QDir>
#include <QProcess>
#include <QTextStream>
#include <QLocale>
#include <QSettings>
#include <QTimer>
#include <QMessageBox>
#include <cclc.h>
#include "cmds.h"
#include "qcclc.h"
#include "grabber.h"
#include "passworddialog.h"
#include "infowin.h"
#include "lockwin.h"

#if defined(Q_WS_WIN32)
# include <windows.h>
# include <winuser.h>
  Q_EXPORT bool qt_win_use_simple_timers;
#endif

static void
onEventCallback(unsigned int cmd, void *data, unsigned int size,
		void *qcclc)
{
  ((Qcclc*)qcclc)->dispatchCommand(cmd, data, size);
}

static void
onDisconnectCallback(void *qcclc)
{
  ((Qcclc*)qcclc)->shutdownNetworking();
  ((Qcclc*)qcclc)->initNetworking();
}

Qcclc::Qcclc(int &argc, char **argv)
  : QApplication(argc, argv),
    m_networking(false), m_active(false), m_stime(-1), m_timeout(-1),
    m_translator(), m_updateStatusTimerId(-1), m_checkEventsTimerId(-1),
    m_tryConnectTimerId(-1)
{
#if defined(Q_OS_WIN32)
  qt_win_use_simple_timers = true;
#endif
#if defined(Q_OS_UNIX)
  QDir::setCurrent(QDir::homePath() + "/.qcclc");
#endif
  loadSettings();
  
  if (!parseCommandLine())
    {
      m_initialized = false;
      printHelp();
      QApplication::exit(1);
    }
  else
    {
      m_initialized = true;
#if defined(Q_OS_WIN32)
      m_translator.load(QString("qcclc_") + m_config.lang);
#else
      m_translator.load(QString("qcclc_") + m_config.lang,
			TRANSLATIONS_DIRECTORY);
#endif
      installTranslator(&m_translator);
      
      m_infoWin = new InfoWin();
      m_grabber = new Grabber(*m_infoWin, m_config.grabberTitle);
      m_lockWin = new LockWin();
      m_passwordDialog = new PasswordDialog();

      loadLockWinSettings();

      connect(m_infoWin, SIGNAL(passwordEdit()), SLOT(passwordEdit()));
      connect(m_infoWin, SIGNAL(endSession()), SLOT(endSession()));
      connect(m_passwordDialog,
	      SIGNAL(setPassword(const QString&, const QString&)),
	      SLOT(setPassword(const QString&, const QString&)));
      connect(this, SIGNAL(timeChanged(int)),
	      m_infoWin, SLOT(setStartTime(int)));
      connect(this, SIGNAL(owedChanged(unsigned int)),
	      m_infoWin, SLOT(setOwed(unsigned int)));
      connect(this, SIGNAL(otherChanged(unsigned int)),
	      m_infoWin, SLOT(setOther(unsigned int)));
      connect(this, SIGNAL(statusChanged(bool)),
	      m_infoWin, SLOT(enableTimeUpdate(bool)));
      connect(this, SIGNAL(connectionStatusChanged(bool)),
	      m_infoWin, SLOT(connectionStatus(bool)));
      connect(this, SIGNAL(enablePassword(bool)),
	      m_infoWin, SLOT(enablePassword(bool)));
      connect(this, SIGNAL(allowMemberLogin(bool)),
	      m_lockWin, SLOT(allowMemberLogin(bool)));
      connect(this, SIGNAL(allowUserLogin(bool)), m_lockWin,
	      SLOT(allowUserLogin(bool)));
      connect(m_lockWin, SIGNAL(userStart()), SLOT(userStart()));
      connect(m_lockWin, SIGNAL(memberStart(int, QString)),
	      SLOT(memberStart(int, QString)));
      connect(m_lockWin, SIGNAL(memberStart(QString, QString)),
	      SLOT(memberStart(QString, QString)));

      m_grabber->show();
      m_infoWin->show();
//      initNetworking();
      QTimer::singleShot(1000, this, SLOT(tryConnect()));
    }
}

Qcclc::~Qcclc()
{
  if (m_initialized)
    {
      shutdownCCLC();
    }
}

void
Qcclc::loadSettings()
{
  QSettings settings("qcclc.cfg", QSettings::IniFormat);

  m_connectionParams.port = settings.value("connection/port", 2999).toUInt();
  m_connectionParams.server =
    settings.value("connection/server", "").toString();
  m_connectionParams.myname =
    settings.value("connection/myname", "").toString();
  m_config.lang =
    settings.value("conf/lang", QLocale::system().name()).toString();
  m_config.grabberTitle =
    settings.value("conf/grabberTitle","Cafe Con Leche").toString();
}

void
Qcclc::loadLockWinSettings()
{
  QSettings settings("qcclc.cfg", QSettings::IniFormat);

  m_lockWin->setBgColor(settings.value("lockwin/bgColor",
				       "#000000").toString());
  m_lockWin->setAuthBoxBgColor(settings.value("lockwin/authBoxBgColor",
					      "#000000").toString());
  m_lockWin->setAuthBoxTextColor(settings.value("lockwin/authBoxTextColor",
						"#ffffff").toString());
  m_lockWin->setAuthBoxTextShadowColor(
      settings.value("lockwin/authBoxTextShadowColor",
		     "#000000").toString());
  m_lockWin->setAuthBoxTitleBgColor(
      settings.value("lockwin/authBoxTitleBgColor",
		     "#64648c").toString());
  m_lockWin->setAuthBoxTitleTextColor(
      settings.value("lockwin/authBoxTitleTextColor",
		     "#ffffff").toString());
  m_lockWin->setAuthBoxTitleTextShadowColor(
      settings.value("lockwin/authBoxTitleTextShadowColor",
		     "#000000").toString());
  m_lockWin->setAuthBoxBorderColor(
      settings.value("lockwin/authBoxBorderColor",
		     "#ffffff").toString());
}

bool
Qcclc::isScreenLocked()
{
  return m_lockWin->isVisible();
}

void
Qcclc::showMessage(const char *message)
{
  if (m_active)
    QMessageBox::information(NULL, tr("Message"), message, QMessageBox::Ok);
}

void
Qcclc::timerEvent(QTimerEvent *timer)
{
  int timerId = timer->timerId();

  if (m_tryConnectTimerId == timerId)
    tryConnect();
  else if (m_checkEventsTimerId == timerId)
    checkEvents();
  else if (m_updateStatusTimerId == timerId)
    updateStatus();
}

void
Qcclc::dispatchCommand(unsigned int cmd, void *data, unsigned int)
{
  unsigned int hdata = 0;

  switch (cmd)
    {
    case CS_STOP:
      stop();
      break;
    case CS_START:
      start();
      break;
    case CS_RESUME:
      resume();
      break;
    case CS_MONITOROFF:
      turnScreenOff();
      break;
    case CS_SHUTDOWN:
      shutdownSystem();
      break;
    case CS_REBOOT:
      rebootSystem();
      break;
    case CS_SETOWED:
      hdata = CCLC_ntohl(*(unsigned int *) data);
      emit owedChanged(hdata);
      break;
    case CS_SETADDITIONAL:
      hdata = CCLC_ntohl(*(unsigned int *) data);
      emit otherChanged(hdata);
      break;
    case CS_SETTIME:
      m_stime = time(NULL) - CCLC_ntohl(*((int *) data));
      emit timeChanged(m_stime);
      break;
    case CS_SETTIMEOUT:
      m_timeout = CCLC_ntohl(*((int *) data));
      emit timeoutChanged(m_timeout);
      break;
    case CS_DISPLAYMESSAGE:
      showMessage((char *) data);
      break;
    case CS_QUITCLIENT:
//      exitProgram();
      break;
    case CS_UNLOCKSCREEN:
      unlockScreen();
      break;
    case CS_LOCKSCREEN:
      lockScreen();
      break;
    case CS_ENABLEPASSWORDEDIT:
      emit enablePassword(bool(CCLC_ntohl((*((int *) data)))));
      break;
    case CS_ALLOWMEMBERLOGIN:
      emit allowMemberLogin(bool(CCLC_ntohl((*((int *) data)))));
      break;
    case CS_ALLOWUSERLOGIN:
      emit allowUserLogin(bool(CCLC_ntohl((*((int *) data)))));
      break;
    }
}

void
Qcclc::initCCLC()
{
  CCLC_init();
  CCLC_set_on_event_callback(onEventCallback, this);
  CCLC_set_on_disconnect_callback(onDisconnectCallback, this);
}

bool
Qcclc::initCCLC(const char *cafile, const char *certfile,
		const char *certpass)
{
  int error;

  initCCLC();

  if (!QDir::current().exists(cafile)
      || !QDir::current().exists(certfile))
    {
      qCritical("[Error] Certificate files not found.");
      return false;
    }
  else if (!CCLC_SSL_init(cafile, certfile, certpass, &error))
    {
      switch (error)
	{
	case CCLC_ERROR_BAD_PASSWORD:
	  qCritical("[Error] Bad certificate password.");
	  return false;
	  break;
	default:
	  qCritical("[Error] Couldn't init SSL.");
	  return false;
	  break;
	}
    }

  return true;
}

void
Qcclc::shutdownCCLC()
{
  CCLC_shutdown();
}

bool
Qcclc::parseCommandLine()
{
  QTextStream err(stderr);
  bool usessl = true;
  char *certpass = NULL;

  int i = 1;

  while (i < argc())
    {
      if (!strcmp(argv()[i],"-host") && ++i < argc())
	m_connectionParams.server = argv()[i];
      else if (!strcmp(argv()[i], "-myname") && ++i < argc())
	m_connectionParams.myname = argv()[i];
      else if (!strcmp(argv()[i], "-port") && ++i < argc())
	m_connectionParams.port = atoi(argv()[i]);
      else if (!strcmp(argv()[i], "-nossl"))
	usessl = false;
      else if (!strcmp(argv()[i], "-certpass") && ++i < argc())
	certpass = argv()[i];
      else
	{
	  qCritical("[Error] Invalid option: '%s'.", argv()[i]);
	  return false;
	}
      ++i;
    }

  if (m_connectionParams.server.isEmpty() || m_connectionParams.myname.isEmpty())
    return false;
  
  if (usessl)
    {
      if (!QDir::current().exists("cert.pem"))
	{
	  qCritical("[Error] File not found: 'cert.pem'.");
	  return false;
	}
      if (!QDir::current().exists("CA.pem"))
	{
	  qCritical("[Error] File not found: 'CA.pem'.");
	  return false;
	}
   
      return initCCLC("CA.pem", "cert.pem", certpass);
    }
  else
    initCCLC();

  return true;
}

void
Qcclc::printHelp()
{
  qWarning(
"Usage: %s -host <server> -myname <myname> [options]\n\n"
"<server>: the server's hostname or IP\n"
"<myname>: the name this client is going to be identified with\n\n"
"Options:\n"
"\t-port <port>: Connect to the specified port (default: 2999).\n"
"\t-nossl: do not use SSL to connect (default: use SSL).\n"
"\t-certpass <password>: password used for the cerfificate (default: none).\n"
"\t                      Ignored if not using SSL (-nossl option enabled).\n"
  , argv()[0]);
}

bool
Qcclc::initNetworking()
{
  int error;
  QTextStream err(stderr);

  if (CCLC_networking_init(m_connectionParams.server.toLatin1().data(),
			   m_connectionParams.port,
			   m_connectionParams.myname.toLatin1().data(),
			   &error))
    {
      CCLC_send_cmd(CC_GETSTATUS,NULL,0);
      m_networking = true;
      m_checkEventsTimerId = startTimer(100);
      m_updateStatusTimerId = startTimer(5000);
      killTimer(m_tryConnectTimerId);
      m_tryConnectTimerId = -1;
      emit connectionStatusChanged(m_networking);
    }
  else
    {
      switch (error)
	{
	case CCLC_ERROR_CONNECT_FAIL:
	  err << "[!] " << tr("Failed to connect with the server.") << "\n";
	  break;
	}

      m_networking = false;
      if (-1 == m_tryConnectTimerId)
	m_tryConnectTimerId = startTimer(5000);
    }

  return m_networking;
}

void
Qcclc::shutdownNetworking()
{
  CCLC_networking_shutdown();
  m_networking = false;
  killTimer(m_checkEventsTimerId);
  killTimer(m_updateStatusTimerId);
  killTimer(m_tryConnectTimerId);
  m_tryConnectTimerId = m_updateStatusTimerId = m_checkEventsTimerId = -1;
  emit connectionStatusChanged(m_networking);
}

void
Qcclc::start()
{
  m_stime = time(NULL);
  m_timeout = 0;
  m_active = true;
  emit timeChanged(m_stime);
  emit owedChanged(0);
  emit otherChanged(0);
  emit timeoutChanged(-1);
  emit statusChanged(m_active);
}

void
Qcclc::stop()
{
  m_active = false;
  emit statusChanged(m_active);
}

void
Qcclc::resume()
{
  m_active = true;
  emit statusChanged(m_active);
}

void
Qcclc::userStart()
{
  CCLC_send_cmd(CC_USERLOGIN, NULL, 0);
}

void
Qcclc::memberStart(int id, QString password)
{
  char val[sizeof(id) + CCLC_MD5_DIGEST_LENGTH * sizeof(unsigned char)];
  unsigned char digest[CCLC_MD5_DIGEST_LENGTH];

  CCLC_MD5((unsigned char *) (password.toLatin1().data()),
	   password.length(), digest);
  ((unsigned int *) val)[0] = CCLC_htonl(id);
  memcpy(((unsigned int *) val) + 1, digest, CCLC_MD5_DIGEST_LENGTH); 
  
  CCLC_send_cmd(CC_MEMBERLOGIN, val, sizeof(val) * sizeof(char));
}

void
Qcclc::memberStart(QString login, QString password)
{
  const char *login_name = login.toLatin1().constData();
  char val[strlen(login_name) * sizeof(char) + 1
	   + CCLC_MD5_DIGEST_LENGTH * sizeof(unsigned char)];
  unsigned char digest[CCLC_MD5_DIGEST_LENGTH];

  memcpy(val, login_name, strlen(login_name) + 1);
  CCLC_MD5((unsigned char *) (password.toLatin1().constData()),
	   password.length(), digest);
  memcpy(val + strlen(val) + 1, digest, CCLC_MD5_DIGEST_LENGTH); 
  CCLC_send_cmd(CC_MEMBERLOGINWITHNAME, val, sizeof(val) * sizeof(char));
}

void
Qcclc::passwordEdit()
{
  m_passwordDialog->showNormal();
}

void
Qcclc::setPassword(const QString &oldpass, const QString &newpass)
{
  unsigned char digests[2 * CCLC_MD5_DIGEST_LENGTH];

  CCLC_MD5((unsigned char *) oldpass.toLatin1().data(),
	   oldpass.length(), digests);
  CCLC_MD5((unsigned char *) newpass.toLatin1().data(),
	   newpass.length(), digests + CCLC_MD5_DIGEST_LENGTH);

  CCLC_send_cmd(CC_SETMEMBERPASSWORD, digests, sizeof(digests));
}

void
Qcclc::endSession()
{
  if (m_active)
    CCLC_send_cmd(CC_USEREXIT, NULL, 0);
  else
    lockScreen();
}

void
Qcclc::lockScreen()
{
  if (m_passwordDialog->isVisible())
    m_passwordDialog->reject();

  m_infoWin->hide();
  m_grabber->hide();
  m_lockWin->lock();
}

void
Qcclc::unlockScreen()
{
  m_lockWin->unlock();
  m_grabber->show();
  m_infoWin->show();
}

void
Qcclc::turnScreenOff()
{
#if defined(Q_WS_X11)
  system("xset dpms force off");
#elif defined(Q_WS_WIN32)
  HWND hWnd = (HWND) m_lockWin->getDC();

  SendMessage(hWnd, WM_SYSCOMMAND, SC_MONITORPOWER, 2);
  m_lockWin->releaseDC((HDC)hWnd);
#elif defined(Q_WS_MAC)
#endif
}

void
Qcclc::shutdownSystem()
{
#if defined(Q_OS_UNIX)
  QProcess::execute("/sbin/shutdown -h now");
#elif defined(Q_OS_WIN32)
  HANDLE hToken;
  TOKEN_PRIVILEGES tkp;

  // Get a token for this process.
  OpenProcessToken(GetCurrentProcess(),
		   TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken);
  // Get the LUID for the shutdown privilege.
  LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid);
  tkp.PrivilegeCount = 1;	// one privilege to set
  tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
  // Get the shutdown privilege for this process.
  AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES) NULL, 0);

  OSVERSIONINFO osversion;

  osversion.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  GetVersionEx(&osversion);
  if (osversion.dwPlatformId == VER_PLATFORM_WIN32_NT)	// NT/2000/XP
    ExitWindowsEx(EWX_POWEROFF | EWX_FORCE,0);
  else
    ExitWindowsEx(EWX_POWEROFF | EWX_FORCE,0);
#endif
}

void
Qcclc::rebootSystem()
{
#if defined(Q_OS_UNIX)
  QProcess::execute("/sbin/shutdown -r now");
#elif defined(Q_OS_WIN32)
  HANDLE hToken;
  TOKEN_PRIVILEGES tkp;

  // Get a token for this process.
  OpenProcessToken(GetCurrentProcess(),
		   TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken);
  // Get the LUID for the shutdown privilege.
  LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid);
  tkp.PrivilegeCount = 1;	// one privilege to set
  tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
  // Get the shutdown privilege for this process.
  AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, (PTOKEN_PRIVILEGES) NULL, 0);

  OSVERSIONINFO osversion;

  osversion.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  GetVersionEx(&osversion);
  if (osversion.dwPlatformId == VER_PLATFORM_WIN32_NT)	// NT/2000/XP
    ExitWindowsEx(EWX_REBOOT | EWX_FORCE,0);
  else
    ExitWindowsEx(EWX_REBOOT | EWX_FORCE,0);
#endif
}

void
Qcclc::updateStatus()
{
  if (!m_networking)
    initNetworking();

  if (m_networking && m_active)
    {
      CCLC_send_cmd(CC_GETOWED, NULL, 0);
      if (0 < m_timeout && (time(NULL) - m_stime) > m_timeout)
	endSession();
    }
}

void
Qcclc::checkEvents()
{
  CCLC_check_events();
}

void
Qcclc::tryConnect()
{
  initNetworking();
}
